from math import sqrt
import configparser

from margot import BackTest
from margot.config import settings

class Strategy(object):
    """
    A trading Strategy.

    Represents a backtested algo, with a recent realised volatility
    and a target volatility for postfolio sizing.

    Args:
        object (BaseAlgo): A margot algrithm.
        recent_vol (float): Recent measured volatility.
        target_vol (float): Target volatility for portfolio contribution.
    """

    def __init__(self, algo, recent_vol, target_vol):   # noqa: D107
        self.algo = algo
        self.recent_vol = recent_vol
        self.target_vol = target_vol

    def vol_size(self):
        return self.recent_vol / self.target_vol


class Portfolio(object):
    """
    A Porfolio of trading Strategies.

    Represents a portfolio of algorithms, each with a measured recent volatility
    and a target volatility. The portfolio has an account size and a target
    volatility for the portfolio as a whole.

    Args:
        object (BaseAlgo): A margot algrithm.
        account_size (int): In dollars, the size of the account.
        target_vol (float): Target volatility for the portfolio as a whole.
    """

    CONFIG_FILE = settings.paths.get('home').joinpath('portfolio.cfg')

    def __init__(self):  # noqa: D107
        # load config
        self.cfg = configparser.ConfigParser()
        self.cfg.read(self.CONFIG_FILE)

    def calibrate(self, cls, algo_name, logger):
        bt = BackTest(algo=cls())
        rets = bt.run(periods=30)
        vol = rets.log_returns.std() * sqrt(252)
        self.cfg.set('margot_calculated', '{}.realised_vol'.format(algo_name), str(vol))
        
        self.cfg.write(self.CONFIG_FILE.open('w'))

        logger.info('Annualised vol at {}'.format(vol))



import logging

import pandas as pd
import numpy as np


logger = logging.getLogger(__name__)


class BackTest(object):
    """Backtest an trading algo that's a descendent of BaseAlgo.

    .. warning:: **BackTest is still a work in progress - it probably doesn't
        even work yet!**

    Attributes:
        algo: A boolean indicating if we like SPAM or not.
        starting_balance: An integer count of the eggs we have laid.
    """

    positions = None
    algo = None
    symbols = set()

    def __init__(self, algo, start_balance=100000):   # noqa: D107
        self.start_balance = start_balance
        self.algo = algo
        self.positions = None
        self.trades = None
        self.returns = pd.DataFrame(
            columns=[
                'simple_returns',
                'log_returns'],
            dtype='float64')

    def calculate_returns(self):
        """Calculate returns.

        Assumes the trade is made the next period after a signal is generated.

        You should construct your MargotDataFrame to be indexed by the trading
        periods (e.g. days).
        """

        # check we have data for the Symbols before we try calculating the
        # returns:
        for column in [col.lower() for col in self.positions.columns]:
            if column not in [sym.lower() for sym in self.algo.data.symbols]:
                raise Exception(
                    'DataFrame does not contain the symbol {} \n Symbols: {}'
                    .format(column, self.algo.data.symbols))

        # TODO: Remove the assumption about simple_returns. i reckon look for the
        # column, then derive simple then log returns.

        # for each symbol that has a position, generate a returns profile

        returns = self.positions.shift()
        for column in returns.columns:
            # log returns np.log(1 + series.pct_change().fillna(0))
            returns.loc[:, column] = self.algo.data.to_pandas(
            ).loc[:, (column, 'returns')] * returns.loc[:, column]

        returns.loc[:, 'returns'] = returns.sum(axis=1)
        returns['log_returns'] = np.log(1 + returns['returns'])
        return returns.replace(np.nan, 0)

    def create_trade_signals_timeseries(self):
        """Create time-series of when position changes occur.

        Return the subset of the positions time-series to indicate positions
            when signals indicate trade should be placed.

        Returns:
            pd.DataFrame: A dataframe of signals when changes to positions are
                suggested.
        """
        # We must start with an unvested position.
        self.positions.iloc[0] = 0
        return self.positions.diff().replace(0, np.nan).dropna(how='all')

    def create_position_timeseries(self, periods):
        """Create Position time-series from signals.

        Runs through all of the backtest data, generating position indicating
        signals.

        Args:
            periods (int): the number of periods to backtest over,
                counted back from the end of the dataset. If no value is
                supplied then the whole dataset is used.

        Returns:
            pd.DataFrame: time-series of Positions
        """
        pos = pd.DataFrame()

        if periods:
            index = self.algo.data.to_pandas(periods).index
        else:
            index = self.algo.data.index

        for ts in index:
            mapper = {}
            pos_list = self.algo.simulate_signal(ts)
            for p in pos_list:
                mapper.update(p.as_map())

            pos = pos.append(pd.DataFrame(mapper, index=[ts]))
        return pos

    def run(self, periods=None):
        """Run the backtest.

        Returns:
            [type]: [description]
        """
        # first calculate the positions time series
        self.positions = self.create_position_timeseries(periods)
        # deduce the trades; simulate entry and exit prices.
        self.trade_signals = self.create_trade_signals_timeseries()

        # Consider the algo being run on 1st day of month e.g. RP
        # can calculate returns daily based on positions held - and
        #   schedule algo runs on e.g. 1st of month.

        # calculate the returns.
        self.returns = self.calculate_returns()
        # caclulate lookback rolling volatility
        # simulate resizing at trading time, according to a target volatility
        # simulate volatility sized returns.
        return self.returns

    def volatility(self, days=30, periods=252):
        """Return a single float value for realised historical volatility.

        TODO: Change the periods parameter to instead examine the data.
        Args:
            days (int, optional): Days to lookback. Defaults to 30.
        """
        # TODO: tail(days) assumes daily.
        return self.returns.log_returns.tail(days).std() * np.sqrt(252)



from margot import Symbol, MargotDataFrame, Ratio, BaseAlgo, Position, BackTest
from margot import finance, cboe, alphavantage as av


# def test_algo():

#     class Index(Symbol):
#         close = cboe.Column(time_series='close')

#     class VXBasis(MargotDataFrame):
#         vixm = Index(symbol='VIX', trading_calendar='NYSE')

#     class MyAlgo(BaseAlgo):
#         data = VXBasis()

#     myalgo = MyAlgo()



import threading
import asyncio
from contextlib import contextmanager

import nest_asyncio

from ib_insync import IB


class IBConnection:
    """
    # usage
    with IBConnection.client() as ib:
        ...
    """
    lock = threading.RLock()
    ib = IB()
    nest_asyncio.apply(ib.client._loop)

    @classmethod
    @contextmanager
    def client(cls):
        try:
            asyncio.get_event_loop()
        except RuntimeError:
            asyncio.set_event_loop(cls.ib.client._loop)

        cls.lock.acquire()
        if not cls.ib.isConnected():
            from config import Config
            port = Config.get('broker.ibkr.port')
            ip_address = Config.get('broker.ibkr.ipaddress')
            cls.ib.connect(
                host=ip_address,
                port=port,
                clientId=10,
                readonly=False)
        cls.lock.release()

        yield cls.ib



# -*- coding: utf-8 -*-
"""
Example Google style docstrings.

Example:
    Examples can be given using either the ``Example`` or ``Examples``
    sections. Sections support any reStructuredText formatting, including
    literal blocks::

        $ python example_google.py

Section breaks are created by resuming unindented text. Section breaks
are also implicitly created anytime a new section starts.

Attributes:
    module_level_variable1 (int): Module level variables may be documented in
        either the ``Attributes`` section of the module docstring, or in an
        inline docstring immediately following the variable.

        Either form is acceptable, but the two should not be mixed. Choose
        one convention to document module level variables and be consistent
        with it.

Todo:
    * For module TODOs
    * You have to also use ``sphinx.ext.todo`` extension

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""
from margot import config  # noqa this is to init the settings

# TODO move these into the sub modules
from margot.data.frames import MargotDataFrame  # noqa
from margot.data.symbols import Symbol  # noqa
from margot.data.ratio import Ratio  # noqa
from margot.data import alphavantage, cboe  # noqa
from margot.data.features import finance  # noqa
from margot.signals.algos import BaseAlgo  # noqa
from margot.signals.backtest import BackTest  # noqa
from margot.signals import Position  # noqa


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import os
from pathlib import Path
import logging
import configparser

from . import settings

CONFIG_FILE = "./etc/margot"


def init():
    # logging
    logger = logging.getLogger("margot config")  # unless worker

    # load config
    cfg = configparser.ConfigParser()
    logger.info("loading config from {}".format(CONFIG_FILE))
    cfg.read(CONFIG_FILE)

    if len(cfg.sections()):
        for section in cfg.sections():
            setattr(settings, section, dict())
            for key, val in cfg.items(section):
                getattr(settings, section)[key] = val
    else:
        logger.warning("config not found at: {}.".format(CONFIG_FILE))
        return

    # find home folder
    base_folder = settings.paths.get("base_folder")
    home = Path.home().joinpath(base_folder)
    logger.debug("using margot_home {}".format(home))

    settings.paths = {
        "home": home,
        "algo_folder": home.joinpath("algos"),
        "log_folder": home.joinpath("logs"),
        "journal_folder": home.joinpath("journal"),
        "venv_folder": home.joinpath("venvs"),
        "cache": home.joinpath("cache"),
    }

    # create directory structure if it doesn't already exist.
    for folder in settings.paths.values():
        if not folder.exists():
            logger.info("creating new directory {}".format(folder))
            folder.mkdir()

    # look for algos
    algo_files = list(settings.paths["algo_folder"].glob("*.cfg"))
    logger.info(
        "found {} cfg files in {}".format(
            len(algo_files), settings.paths["algo_folder"]
        )
    )

    # create a dict to hold the algo configs (ConfigParser objects)
    settings.algos = dict()

    for algo_file in algo_files:
        logger.debug("parsing {}".format(algo_file))
        algo_config = configparser.ConfigParser()
        algo_config.read(algo_file)
        algo_name = algo_config.get("python", "file")

        settings.algos[algo_name] = dict()

        if len(algo_config.sections()):
            for section in algo_config.sections():
                settings.algos[algo_name][section] = dict()
                for key, val in algo_config.items(section):
                    settings.algos[algo_name][section][key] = val
        else:
            logger.warning("config not found at: {}.".format(CONFIG_FILE))
            return

    # lets collect up the os environment variables here too.
    if not hasattr(settings, "env"):
        settings.env = dict()

    for key in os.environ:
        settings.env[key] = os.environ[key]


if not settings.INITED:
    init()





class Ratio(object):
    """Ratio of two series.

    Ratio = numerator / denominator of two time-series.

    Example::

        my_ratio = Ratio(numerator=symbol_1.adj_close,
                         denominator=symbol_2.adj_close,
                         label='s1_s2_ratio')

    Args:
        numerator (str): the series to numerate the ratio
        denominator (str): the series to denominate the ratio
        label (str): give it a name for your dataframe. e.g. current_ratio
    """

    def __init__(self, numerator, denominator, label, **kwargs):    # noqa: D107
        self.numerator = numerator
        self.denominator = denominator
        self._series = None
        self.label = label
        self.__dict__.update(kwargs)

        if not self.numerator and self.denominator:
            raise AttributeError(
                "You must supply both a numerator and denominator")

        self.series = self.make_series()

    def make_series(self):
        return self.numerator.series.divide(self.denominator.series)

    @property
    def series(self):
        return self._series.rename(self.label)

    @series.setter
    def series(self, series):
        self._series = series

    def to_pandas(self):
        return self.series.to_frame()

    def simulate(self, when=None):
        self.series = self.make_series()[:when]

    @property
    def latest(self):
        """Return the latest value in this series."""
        return self.series.tail(1)[0]



import os
from margot import Symbol, MargotDataFrame, Ratio, BaseAlgo, Position, BackTest
from margot import finance, cboe, alphavantage as av


def test_symbol():
    class Equity(Symbol):
        adjusted_close = av.DailyAdjusted(time_series="adjusted_close")
        volume = av.DailyAdjusted(time_series="volume")

        simple_returns = finance.SimpleReturns(column="adjusted_close")
        log_returns = finance.LogReturns(column="adjusted_close")
        realised_vol = finance.RealisedVolatility(
            column="log_returns", window=30
        )

        upper_band = finance.UpperBollingerBand(
            column="adjusted_close", window=20, width=2.0
        )
        sma20 = finance.SimpleMovingAverage(column="adjusted_close", window=20)
        lower_band = finance.LowerBollingerBand(
            column="adjusted_close", window=20, width=2.0
        )

    env = {"DATA_CACHE": os.path.join(os.getcwd(), "data")}
    spy = Equity(symbol="SPY", trading_calendar="NYSE")


# def test_frame():
#     class Index(Symbol):
#         close = cboe.Column(time_series="close")

#     class VXBasis(MargotDataFrame):
#         vixm = Index(symbol="VIX3M", trading_calendar="NYSE")
#         vix3m = Index(symbol="VIX", trading_calendar="NYSE")
#         vx_basis = Ratio(
#             numerator=vixm.close,
#             denominator=vix3m.close,
#             label="vx_basis_ratio",
#         )

#     vxbasis = VXBasis()
#     vxbasis.to_pandas()
#     vxbasis.vx_basis.latest
#     vxbasis.vx_basis.to_pandas()


def test_constructors():
    class Index(Symbol):
        adjusted_close = av.DailyAdjusted(time_series="adjusted_close")

    spy = Index(symbol="SPY", trading_calendar="NYSE")
    vtwo = Index(symbol="VTWO", trading_calendar="NYSE")

    assert (
        spy.to_pandas().tail()["SPY", "adjusted_close"].sum()
        != vtwo.to_pandas().tail()["VTWO", "adjusted_close"].sum()
    )


def test_finance_features():
    class Index(Symbol):
        adjusted_close = av.DailyAdjusted(time_series="adjusted_close")

        simple_returns = finance.SimpleReturns(column="adjusted_close")
        log_returns = finance.LogReturns(column="adjusted_close")
        real_vol = finance.RealisedVolatility(column="log_returns", window=20)
        sma = finance.SimpleMovingAverage(column="adjusted_close", window=10)
        upper = finance.UpperBollingerBand(
            column="adjusted_close", window=20, width=2.0
        )
        lower = finance.LowerBollingerBand(
            column="adjusted_close", window=20, width=2.0
        )

    spy = Index(symbol="SPY", trading_calendar="NYSE")

    spy.to_pandas()



import asyncio
import logging
import json
from pathlib import Path

from margot.config import settings

logger = logging.getLogger('margot')


class MargotJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, Path):
            return str(obj)


@asyncio.coroutine
def handle_message(reader, writer):
    data = yield from reader.readline()
    msg = data.decode()
    logger.info('received: {}'.format(msg))
    # it now goes to the portfolio manager
    tokens = msg.split()

    if tokens[0] == 'GETALGO':
        reply = json.dumps(
            settings.algos.get(
                tokens[1],
                'NONE'),
            cls=MargotJSONEncoder)
        writer.write(reply.encode())
        writer.close()
        logger.debug('Sent reply: {}'.format(reply))


def init(logger):
    logger.info('starting server')
    loop = asyncio.get_event_loop()
    coro = asyncio.start_unix_server(
        handle_message,
        path=settings.sys.get('socket'),
        loop=loop
    )
    return loop.run_until_complete(coro)



import os
import logging
from datetime import datetime
from pathlib import Path

import pytz
import pandas as pd

from margot.config import settings

logger = logging.getLogger(__name__)


class DailyMixin(object):

    @property
    def stale(self):
        """Check if we think there might be new data, and if so - update.
        """
        now = pd.Timestamp(datetime.now(tz=pytz.UTC))
        if self._full_series.index.max().date(
        ) < self.trading_calendar.previous_close(now).date():
            return True


class BaseColumn(object):
    """
    BaseColumn is the super class for implementing Columns.

    Generally, you will only need to extend BaseColumn if you
    are implementing a new data provider.

    A Column represents a single time series of a symbol.

    Examples of commonly used time-series are adjusted_close,
    open, highh, low close, volume. However columns can also
    be used to represent fundamental time-series, or time-
    series from alternative sources.

    To implement a new type of Column, you must implement the
    'fetch' method.

    Example::

        class MyDataProvider(BaseColumn):

            def fetch(self, symbol):
                df = get_my_dataframe(symbol)
                return self.clean(df)

    Optionally, you may also need to perform additional cleaning
    of the data, you can do this by extending the clean() method.
    Don't forget to call super().clean(df).

    Example::

        class MyDataProvider(BaseColumn):

            def clean(self, df):
                df = df.rename(mapper={
                    'Open': 'open',
                    'High': 'high',
                    'Low': 'low',
                    'Close': 'close',
                }, axis='columns')

                return super().clean(df)

    When using an implementation of a subclass of BaseColumn, users are
    expected to at least specify the time_series that they want to access.

    Args:
        time_series (str): the name of the time_series that will be returned.
    """

    INITED = False

    def __init__(self, time_series: str, *args, **kwargs):  # noqa: D107
        self.time_series = time_series
        self._full_series = None
        self._series = None

    def get_label(self):
        """Return the label for this column."""
        return self.series.name

    def clone(self):
        """Return a new instance of oneself."""
        return self.__class__(self.time_series)

    def setup(self, symbol: str, trading_calendar):
        """
        Called by Symbol after it is instantiated.

        Ensures that this column knows the context in which it operates,
        including what Symbol we are using and what the trading_calendar is.
        """
        self.symbol = symbol
        self.trading_calendar = trading_calendar

        # TODO: File names should be managed in a central configuration
        data_cache = settings.paths['cache']
        Path(data_cache).mkdir(parents=True, exist_ok=True)

        self.hdf5_file = os.path.join(
            data_cache, '{}.hdf5'.format(self.symbol))

    @classmethod
    def clean(cls, df):
        """Clean the data."""
        df = df.sort_index()
        # make tz aware if not already
        if not isinstance(df.index.dtype, pd.DatetimeTZDtype):
            df = df.tz_localize(pytz.UTC)

        return df

    def load_or_fetch_series(self, symbol: str):
        """
        Load of fetch the Dataframe, return the series.

        In order to return the time-series, first determine if we
        have it and can return it, or if we need to fetch it.

        Args:
            symbol (str): the name of the symbol to fetch.

        Returns:
            pd.Series: time-series of the column
        """
        try:
            df = self.load(symbol)
        except (KeyError, FileNotFoundError):
            df = self.refresh()
        return df[self.time_series]

    def refresh(self):
        """
        Refresh the data from the source.

        Returns:
            pd.DataFrame: the whole dataframe (cleaned)
        """
        df = self.fetch(self.symbol)
        df = self.clean(df)
        self.save(df, self.symbol)
        return df

    def fetch(self, symbol: str):  # noqa: D102
        raise NotImplementedError(
            'This is implementation specific to the data provider.')

    def load(self, symbol: str):
        """Load it."""
        return pd.read_hdf(
            self.hdf5_file,
            key=symbol)

    def save(self, df, symbol):
        """Save it."""
        df.to_hdf(self.hdf5_file, key=symbol)

    @property
    def series(self):
        """
        Get the data series as a pandas series.

        Returns:
            pd.Series: time series of the field
        """
        if self._full_series is None:
            self._full_series = self.load_or_fetch_series(self.symbol)
            if self.stale:
                self.refresh()
                self.load(self.symbol)
            self._series = self._full_series.copy()

        self.INITED = True
        return self._series

    def simulate(self, when):
        """
        Simulate a time in history.

        Args:
            when (datetime): (optional) used when
                simulating historical data, typically
                using margot.backtest.
        """
        self._series = self._full_series[:when]

    @property
    def latest(self):
        """Return the latest value in this series."""
        return self.series.tail(1)[0]



import os
import logging

from alpha_vantage.timeseries import TimeSeries

from margot.data.columns import BaseColumn, DailyMixin

from margot.config import settings

logger = logging.getLogger(__name__)


class DailyAdjusted(BaseColumn, DailyMixin):
    """A daily time series from AlphaVantage.

    Example::

        from margot import alphavantage as av

        volume = av.DailyAdjusted(time_series='adjusted_close')

    Args:
        time_series (str): the name of the time-series that will be returned.
            Can be one of: 'open', 'high', 'low', 'close', 'adjusted_close',
            'volume', 'dividend_amount' or 'split_coefficient'.
    """

    def clean(self, df):
        """
        Clean the dataframe.

        Alphavantage has odd column names, so we'll fix those.
        """
        df = df.rename(mapper={
            '1. open': 'open',
            '2. high': 'high',
            '3. low': 'low',
            '4. close': 'close',
            '5. adjusted close': 'adjusted_close',
            '6. volume': 'volume',
            '7. dividend amount': 'divident_amount',
            '8. split coefficient': 'split_coefficient'
        }, axis='columns')
        return super().clean(df)

    def fetch(self, symbol: str):
        """
        Fetch from remote - this could be the only service specific thing.

        Args:
            symbol (str): the name of the symbol to fetch

        """
        logger.info('fetching ({}) from alphavantage'.format(symbol))
        ts = TimeSeries(
            key=settings.env['ALPHAVANTAGE_API_KEY'],
            output_format='pandas')
        df, _ = ts.get_daily_adjusted(symbol, outputsize='full')
        return self.clean(df)



import logging
import ftx

class FTXObject():

    def __init__(self, token):
        for key in token:
            setattr(self, key, token[key])


class LeveragedToken(FTXObject):
    
    def calc_pending_rebalance(self):
        """
        Calculate pending rebalance.
        """
        desired_position = float(self.leverage) * \
        self.totalNav / self.underlyingMark

        current_position = self.positionPerShare * \
        self.outstanding

        return desired_position - current_position


class Future(FTXObject):

    def __init__(self, *args, **kwargs):
        self._lt_list = None
        super(Future, self).__init__(*args, **kwargs)

    @property
    def hourly_volume(self):
        return self.volume / 24

    @property
    def hourly_volume_usd(self):
        return self.volumeUsd24h / 24

    @property
    def lt_list(self):
        return self._lt_list

    def set_lt_list(self, lt_list):
        self._lt_list = lt_list

    @property
    def market_impact(self):
        """
        Basic metric for market impact.
        """
        return self.rebal_size / self.hourly_volume

    @property
    def pending_rebal_usd(self):
        return self.rebal_size * self.mark

    @property
    def rebal_size(self):
        pending_rebalance = 0.0
        for lt in self.lt_list:
            pending_rebalance += lt.calc_pending_rebalance()
        return pending_rebalance



class FTXStrategy():

    BUY = LONG = 'buy'
    SELL = SHORT = 'sell'

    perps = list()
    futures = list()
    leveraged_tokens = list()

    def __init__(self, api_key, api_secret, subaccount, debug) -> None:

        self.client = ftx.FtxClient(
            api_key=api_key, 
            api_secret=api_secret, 
            subaccount_name=subaccount)

        for token in self.client.list_lts():
            lt = LeveragedToken(token)
            self.leveraged_tokens.append(lt)

        for future in self.client.get_futures():
            lt_list = [
                lt for lt in self.leveraged_tokens
                    if lt.underlying == future['name']
                ]

            f = Future(future)
            f.set_lt_list(lt_list)
            self.futures.append(f)

        self.perps = [future for future in self.futures if future.perpetual]

        self.debug=debug

    def place_order(self, market, side, size_usd):
        latest = self.client.get_future(market)

        if side == self.BUY:
            limit = latest['ask'] * 1.0005
        else:
            limit = latest['bid'] * 0.9995

        size = size_usd / limit

        logging.info('debug={}, '\
            'market="{}", ' \
            'side="{}", ' \
            'price={:,.4f}, ' \
            'size={:,.4f}'.format(self.debug,
                                    market, 
                                    side, 
                                    limit, 
                                    size))

        if not self.debug:
            try:
                self.client.place_order(
                    market=market,
                    side=side, 
                    price=limit,
                    size=size,
                    type='limit')
            except Exception as e:
                print(e)


    def close_position(self, position):
        latest = self.client.get_future(position['future'])

        if position['side'] == self.SHORT:
            desired_side = self.BUY
        else:
            desired_side = self.SELL

        if desired_side == self.BUY:
            limit = latest['ask'] * 1.0010
        else:
            limit = latest['bid'] * 0.9990

        logging.info('debug={}, '\
                'market="{}", ' \
                'side="{}", ' \
                'price={:,.4f}, ' \
                'size={:,.4f}'.format(
                    self.debug,
                    position['future'], 
                    desired_side, 
                    limit, 
                    position['size']))

        if not self.debug:
            try:
                self.client.place_order(
                    market=position['future'],
                    side=desired_side, 
                    price=limit,
                    size=position['size'],
                    type='limit',
                    reduce_only=True)
            except Exception as e:
                print(e)


    def close_all_positions(self):
        """
        While closing the position, I want to be more agressive
        to get out before the reversion kicks in.
        """
        for position in self.client.get_positions():
            if position['size'] != 0.0:
                self.close_position(position)
    
        print('Closed all positions.')


from datetime import datetime

import pytz

from margot import Symbol, MargotDataFrame, Ratio, BaseAlgo, Position, BackTest
from margot import finance, cboe, alphavantage as av


# class Index(Symbol):
#     close = cboe.Column(time_series='close')
#     sma = finance.SimpleMovingAverage(column='close',
#                                       window=3)


# class Equity(Symbol):
#     close = av.DailyAdjusted(time_series='adjusted_close')
#     simple_returns = finance.SimpleReturns(column='close')


# class VXBasisDF(MargotDataFrame):
#     vix = Index('VIX', trading_calendar='NYSE')
#     vix3m = Index('VIX3M', trading_calendar='NYSE')
#     ratio = Ratio(numerator=vix.close,
#                   denominator=vix3m.close,
#                   label='vxbasis')
#     ziv = Equity('ZIV', trading_calendar='NYSE')


# class VXBasisTrade(BaseAlgo):
#     data = VXBasisDF()

#     def signal(self):
#         if self.data.ratio.latest <= 1.0 and \
#                 self.data.vix.close.latest <= self.data.vix.sma.latest:
#             return [Position('ZIV', 1.0, self.MOC)]
#         else:
#             return [Position('ZIV', 0.0, self.MOC)]


# def test_simulation():
#     vxa = VXBasisDF()
#     vxb = vxa.to_pandas().copy()
#     when = datetime(2020, 1, 1, tzinfo=pytz.UTC)
#     vxa.simulate(when)
#     assert(vxb[:when].equals(vxa.to_pandas()))


# def test_backtest():
#     vx = VXBasisTrade()
#     bt = BackTest(vx)
#     bt.run(periods=100)

def dummy_test():
    assert True


import pandas as pd

from margot.data.columns import BaseColumn


class BaseFeature(object):

    INITED = False
    label = None

    def __init__(self, column: str, *args, **kwargs):  # noqa: D107
        self.column = column
        self.__dict__.update(kwargs)
        self.kwargs = kwargs
        self.args = args
        self._series = None
        self.base_column = None

    def clone(self):
        """Return a new instance of oneself."""
        return self.__class__(self.column, *self.args, **self.kwargs)

    def set_column(self, base_column: BaseColumn):
        self.base_column = base_column

    def get_column_name(self) -> str:
        return self.column

    @property
    def series(self):
        if self._series is None:
            # TODO - consider set_column - used after cloning.
            series = self.feature(self.base_column.series)
            self._series = series.rename(self.get_label())
        return self._series

    @series.setter
    def set_series(self, series):
        self._series = series

    def get_label(self):
        """
        Return the label for this feature.

        Override this to customise.

        Returns:
            str: the label to be used in the pandas column.
        """
        return self.label

    def feature(self, series: pd.Series):  # noqa: D102
        raise NotImplementedError("please implement the feature")

    def simulate(self, when):
        """Recalculate the feature, typically in a simulation."""
        self._series = None
        self._series = self.series[:when]

    @property
    def latest(self):
        """Return the latest value in this series."""
        return self.series.tail(1)[0]



import socket
import importlib
import inspect

from margot import BaseAlgo
from margot.config import settings


def send_message(msg):
    # post the signal to the manager... let him deal with it.
    try:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(settings.sys.get('socket'))
        sock.sendall(msg.encode())
        sock.close()
        logger.info('Sent to manager.')
    except FileNotFoundError:
        logger.error('Unable to connect to server. Is it running?')
        raise OSError(
            'Unable to connect to {}'.format(
                settings.sys.get('socket')))


def init(algo_file, config, logger):
    logger.debug('initialising worker for {}'.format(algo_file))
    for name, cls in inspect.getmembers(
            importlib.import_module(algo_file),
            inspect.isclass):

        if issubclass(cls, BaseAlgo) and cls != BaseAlgo:
            logger.debug('Found algo {}'.format(cls))
            algo = cls()
            signal = algo.signal()
            logger.info('signal is: {}'.format(signal))
            message = 'SIGNAL {} | {} \n'.format(name, str(signal))
            send_message(message)



import configparser
import subprocess

from margot.config import settings


def add_job(algo, logger, sched, verbose=True):
    cmd = 'cd {} && {}/bin/python -m margot -v -w {}'.format(
        algo['python']['working_dir'],
        algo['venv'],
        algo['python']['file']
    )

    logger.info('adding job: {}'.format(cmd))
    trigger = algo['schedule']['trigger']

    if trigger == 'cron':
        return sched.add_job(
            subprocess.run,
            args=[cmd],
            kwargs={'shell': True},
            trigger='cron',
            day_of_week=algo['schedule'].get('day_of_week'),
            hour=algo['schedule'].get('hour'),
            minute=algo['schedule'].get('minute'),
            second=algo['schedule'].get('second'),
            jitter=int(algo['schedule'].get('jitter'))
        )

    if trigger == 'interval':
        return sched.add_job(
            subprocess.run,
            args=[cmd],
            kwargs={'shell': True},
            trigger='interval',
            minutes=algo['schedule'].get('minutes'),
        )


def load_algo(algo, logger, sched):
    """
    Load an algo from its cfg file.
    """

    # does the venv exist?
    algo['venv'] = settings.paths['venv_folder'].joinpath(
        algo['python']['venv'])

    if not algo['venv'].exists():
        logger.info('creating venv {}'.format(algo['venv']))
        algo['venv'].mkdir()
        python3 = algo['python'].get('interpreter', 'python3')
        logger.info(subprocess.run(
            '{} -m venv {}'.format(python3, algo['venv']),
            shell=True))

        # and install requirements
        for req, ver in algo['requirements'].items():
            logger.debug('installing requirement {} {}'.format(req, ver))
            logger.info(subprocess.run(
                '{}/bin/pip install {} {}'.format(algo['venv'], req, ver),
                shell=True))

    add_job(algo, logger, sched)

    # register algo in registry



import socket
import json

from math import sqrt

from margot.config import settings
from margot import BaseAlgo
from margot.portfolio import Portfolio

import importlib
import inspect


def ipc_request(msg, logger):
    # requst the manager for something
    try:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(settings.sys.get('socket'))
        sock.sendall(msg.encode())
        logger.debug('Message sent: {}'.format(msg))
        data = sock.recv(4096)
        sock.close()
        reply = data.decode()
        logger.debug('Received {}'.format(reply))
        return json.loads(reply)

    except FileNotFoundError:
        logger.error('Unable to connect to server. Is it running?')
        raise OSError(
            'Unable to connect to {}'.format(
                settings.sys.get('socket')))


def init(algo_name, settings, logger):
    algo = ipc_request('GETALGO {} \n'.format(algo_name), logger)
    logger.info('Inspecting {}'.format(algo.get('algorithm').get('name')))

    for name, cls in inspect.getmembers(
            importlib.import_module(algo.get('python').get('file')),
            inspect.isclass):
        logger.debug('Inspecting {}'.format(cls))
        if issubclass(cls, BaseAlgo) and cls != BaseAlgo:
            logger.info('Found algo class {}'.format(cls))
            pf = Portfolio()
            pf.calibrate(cls, algo_name, logger)

    # need to store the backtests for each algo so that we have volatility etc.



# todo

# Live trading

# 1. Make an apscheduler thing that can pass jobs to ibinsync
# 2. Make a way to configure which algos are run when - this should be loaded via config file initially.
# 3. Run those algos at those times.

# IB should be pluggable (i.e. we could use something else)



from inspect import getmembers

import pandas as pd

from margot.data.features import BaseFeature
from margot.data.symbols import Symbol
from margot.data.ratio import Ratio


class MargotDataFrame(object):
    """A MargotDataFrame brings together symbols, columns, features and ratios.

    Example::

        class Equity(Symbol):
            adj_close = av.Column(function='historical_daily_adjusted',
                                  time_series='adjusted_close')
            log_returns = finance.LogReturns(column='adj_close')
            realised_vol = finance.RealisedVolatility(column='log_returns',
                                                      window=30)

        class ExampleDF(MargotDataFrame):
            spy = Equity(symbol='SPY', trading_calendar='NYSE')
            vtwo = Equity(symbol='VTWO', trading_calendar='NYSE')
            spy_russ_ratio = Ratio(numerator=spy.adj_close,
                                   denominator=vtwo.adj_close,
                                   label='spy_russ')

        mydf = ExampleDF()

    """

    def __init__(self):   # noqa: D107
        self.symbols = [
            name for name,
            ref in getmembers(self, lambda m: isinstance(m, Symbol))]

        self.features = [
            name for name,
            ref in getmembers(self, lambda m: isinstance(m, BaseFeature))]

        self.ratios = [
            name for name,
            ref in getmembers(self, lambda m: isinstance(m, Ratio))]
        super().__init__()

    def to_pandas(self, periods: int = None, dropna=True) -> pd.DataFrame:
        """Return a pandas Dataframe representing this MargotDataFrame.

        Args:
            periods (int, optional): only return the tail n periods.

        Returns:
            pd.DataFrame: a Pandas dataframe representing all data from
                the MargotDataFrame
        """
        # Get the elements one at a time, to pandas them and ensemble.
        if len(self.symbols) == 1:
            df1 = self.symbols[0].to_pandas()
        elif len(self.symbols) > 1:
            df1 = pd.concat([getattr(self, name).to_pandas()
                             for name in self.symbols], axis=1)
        else:
            df1 = pd.DataFrame()

        df2 = pd.DataFrame({('margot', name): getattr(self, name).series
                            for name in self.ratios + self.features})

        df = pd.concat([df1, df2], axis=1)

        if dropna:
            df = df.dropna()

        if periods:
            df = df.tail(periods)

        return df

    def refresh(self):
        """Refresh all Symbols in this DataFrame."""
        for member in self.symbols:
            getattr(self, member).refresh()
        # TODO what about ratios?

    @property
    def start_date(self):
        """First Timestamp of the time-series index.

        Returns:
            Timestamp: a pandas timestamp.
        """
        return self.to_pandas().index.min()

    @property
    def end_date(self):
        """Last Timestamp value of the time-series index.

        Returns:
            Timestamp: a pandas timestamp.
        """
        return self.to_pandas().index.max()

    @property
    def index(self):
        """Return the time-series index.

        Returns:
            pd.Index: a pandas timeseries index.
        """
        return self.to_pandas().index

    @property
    def when(self):
        return self._when

    @when.setter
    def set_when(self, when):  # noqa: D102
        self._when = when

    def simulate(self, when):
        """Create a dataframe simulating a datetime in history.

        Used for backtesting to simplify the writing of trading
        algorithms. After simulating a historical datetime, it is
        not possible to go back to the future.

        Args:
            when (tz_aware datetime or pd.Timestamp): when to go back to.
        """
        self._when = when

        for symbol in self.symbols:
            getattr(self, symbol).simulate(when)

        for feature in self.features:
            getattr(self, feature).simulate(when)

        for ratio in self.ratios:
            getattr(self, ratio).simulate(when)

    def end_simulation(self):
        self._when = None

        for symbol in self.symbols:
            getattr(self, symbol).simulate()

        for feature in self.features:
            getattr(self, feature).simulate()

        for ratio in self.ratios:
            getattr(self, ratio).simulate()



import logging
from datetime import datetime

import pandas as pd

from trading_calendars import get_calendar

import pytz

from margot.signals.order_types import MOC, MKT, LMT
from margot import MargotDataFrame

logger = logging.getLogger(__name__)


class BaseAlgo(object):
    """A base class to inherit when implementing your trading algorithm.

    You should at least implement signal() which is the output
    of a trading algorithm.

    Args:
        env (dict): a dictionary of environment variables, e.g. API keys.
            Overrides anything provided in sys env.
        market (str): The ISO code for the market we will use.

    Raises:
        ValueError: the attribute, 'data' must be a reference to a MargotDataFrame.
        NotImplementedError: If your subclass does not implement signal(), you will
            receive a NotImplementedError.

    """

    MOC = MOC
    MKT = MKT
    LMT = LMT

    MONDAY = 'MON'
    TUESDAY = 'TUE'
    WEDNESDAY = 'WED'
    THURSDAY = 'THU'
    FRIDAY = 'FRI'
    SATURDAY = 'SAT'
    SUNDAY = 'SUN'

    data = None

    def __init__(self, market='XNYS'):  # noqa: D107
        self.market = get_calendar(market)
        self.when = None
        if not isinstance(self.data, MargotDataFrame):
            raise ValueError('Please set data to reference a MargotDataFrame')

    def weekday(self, dt):
        """Return a human readable three letter day of week.

        Convert the Python integer representation of day of week into a string.

        e.g::

            0: 'MON' (also known as self.MONDAY)

        .. note::
            You should always use the built in constants when passing days of
            the week. e.g. self.MONDAY, self.TUESDAY, ... these map to the three
            charater strings.


        Args:
            dt (datetime or pd.Timestamp): The datetime to check

        Returns:
            str: One of; 'MON', 'TUE', 'WED', 'THU', 'FRI', 'SAT', 'SUN'
        """
        days = {
            0: self.MONDAY,
            1: self.TUESDAY,
            2: self.WEDNESDAY,
            3: self.THURSDAY,
            4: self.FRIDAY,
            5: self.SATURDAY,
            6: self.SUNDAY
        }
        return days.get(dt.weekday())

    @property
    def next_close(self):
        """Return a UTC pd.Timestamp of the next close of trading session.

        Returns:
            pd.Timestamp: Timestamp of the next close of cash session in UTC.
        """
        return self.market.next_close(
            getattr(self, 'when',
                    pd.Timestamp(datetime.now(tz=pytz.UTC)))
        )

    def signal(self) -> list:
        """Return a list of Position objects for a given datetime."""
        raise NotImplementedError("You must implement signal")

    def simulate_signal(self, when: datetime):
        """Simulate a signal from a point in time.

        Stores the original MargotDataFrame referenced by self.data
        on a temporary reference so that the data attribute can be
        used by signal() to calculate positions at a point in history.

        After running signal(), the full dataframe is re-referenced
        at self.data.

        Args:
            when (datetime): when in history to go back to

        Returns:
            list: a list of Position objects.
        """
        self.data.simulate(when)
        self.when = when
        positions = self.signal()
        self.data.end_simulation()
        return positions



import json
from google.cloud import secretmanager


def gcp_get_secret(project_id, secret_id):
    """
    Access the payload for the given secret version if one exists. The version
    can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
    """

    # Import the Secret Manager client library.
    from google.cloud import secretmanager

    # Create the Secret Manager client.
    client = secretmanager.SecretManagerServiceClient()

    # Build the resource name of the secret version.
    name = f"projects/{project_id}/secrets/{secret_id}/versions/latest"

    # Access the secret version.
    response = client.access_secret_version(request={"name": name})

    # Print the secret payload.
    #
    # WARNING: Do not print the secret in a production environment - this
    # snippet is showing how to access the secret material.

    return json.loads(response.payload.data.decode("UTF-8"))



from setuptools import setup, find_packages
import versioneer

with open("README.md") as f:
    long_description = f.read()

setup(
    name="margot",
    packages=find_packages(),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="apache-2.0",
    description="An algorithmic trading framework for PyData.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Rich Atkinson",
    author_email="rich@airteam.com.au",
    url="https://github.com/pymargot/margot",
    download_url="",
    keywords=["quant", "trading", "systematic"],
    install_requires=[
        "numpy",
        "pandas",
        "scipy",
        "pyfolio",
        "exchange_calendars",
        "versioneer",
        "pytz",
        "alpha_vantage",
        "tables",
        "ftx",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from pathlib import Path
from apscheduler.schedulers.asyncio import AsyncIOScheduler

# TODO move all this to a global config registry
LOCKFILE = '/tmp/.margot-lockfile'


def init(logger):
    if not Path(LOCKFILE).exists():
        scheduler = AsyncIOScheduler()
        scheduler.start()
        # We use a lockfile to prevent multiple servers running
        Path(LOCKFILE).touch()
        return scheduler
    else:
        logger.error('A manager is already running')
        raise Exception('A manager is already running')


def close(logger):
    logger.info('Shutting down the scheduler')
    Path(LOCKFILE).unlink()



import asyncio

from .algos import load_algo
from . import scheduler, server
from margot.config import settings


def init(config, logger):
    try:
        # the scheduler runs algos on their schedule
        sched = scheduler.init(logger)

        # the server receives trading messages from algos
        server.init(logger)

        for algo in settings.algos.keys():
            load_algo(settings.algos[algo], logger, sched)
            # connect to brokers
            # load portfolio definition

        logger.info('running forever')
        asyncio.get_event_loop().run_forever()
    except (KeyboardInterrupt, SystemExit):
        logger.info('manager has been killed')
        scheduler.close(logger)
        pass
    except BaseException:
        logger.info('guru meditatiton - cleaning up lockfile')
        scheduler.close(logger)
        raise



#!/usr/bin/env python3

import sys
import argparse
import configparser
import logging

from ._version import get_versions

from margot import config  # noqa this is to init the settings
from margot.config import settings

from margot.trade import manager, worker, calibrate
from margot.trade.algos import load_algo

import warnings
# warnings.filterwarnings('error')


def main(args):

    # logging
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(format="%(levelname)s: %(message)s", level=loglevel)
    logger = logging.getLogger('margot')

    if args.config:
        # reload config from new file
        settings.CONFIG_FILE = args.config
        config.init()

    if args.version:
        print('margot version {}'.format(get_versions()['version']))
        return

    if args.calibrate:
        calibrate.init(args.calibrate, settings, logger)

    elif args.worker:
        logger = logging.getLogger(args.worker)
        worker.init(args.worker, settings, logger)

    elif args.server:
        logger = logging.getLogger('margot')
        logger.info(
            'starting margot manager (version {})'.format(
                get_versions()['version']))
        manager.init(settings, logger)

    else:
        print('--help for help.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog='margot',
        description="Schedule and execute trading strategies with champagne and style.",
        epilog="As an alternative, params can be placed in a file, one per line, and specified as '%(prog)s @params.conf'.",
        fromfile_prefix_chars='@')

    parser.add_argument(
        "--version",
        help="Print margot version then exit.",
        action="store_true")

    parser.add_argument(
        "-v",
        "--verbose",
        help="Increase output verbosity for debugging",
        action="store_true")

    parser.add_argument(
        "--config",
        metavar="inifile",
        help="Specify a configuration file to use, instead of the default.",)

    parser.add_argument(
        "-c",
        "--calibrate",
        metavar="algofile",
        help="Calibrate an algorithm by measuring it's recent realised volatility",)

    parser.add_argument(
        "-w",
        "--worker",
        help="A worker runs an algorithm, usually invoked by the manager.",
        metavar="algofile",)

    parser.add_argument(
        "-s",
        "--server",
        help="Server & manager eventloop. Schedules workers and manages trade execution.",
        action="store_true")

    main(parser.parse_args())
